from __future__ import annotations

import math
from collections import deque
from collections.abc import Callable, Iterable, Iterator
from functools import partial
from itertools import count, groupby
from typing import Any, Literal, TypeGuard, TypeVar, overload

__all__ = [
    "map",
    "filter",
    "for_each",
    "for_batch",
    "batch",
    "branch",
]

T = TypeVar("T")
V = TypeVar("V")
U = TypeVar("U")
W = TypeVar("W")
X = TypeVar("X")


raw_filter = filter


Step = Callable[[Iterable[T]], Iterable[V]]


def map(step: Callable[[V], W]) -> Step[V, W]:
    def f(data: Iterable[V]) -> Iterable[W]:
        for item in data:
            yield step(item)

    return f


def for_each(step: Callable[[V], Any]) -> Step[V, V]:
    def f(data: Iterable[V]) -> Iterable[V]:
        for item in data:
            step(item)
            yield item

    return f


def for_batch(step: Callable[[list[V]], Any], batch_size: int) -> Step[V, V]:
    def f(data: Iterable[V]) -> Iterable[V]:
        for _, batch_iterator in groupby(
            zip(data, count()),
            key=lambda x: math.floor(x[1] / batch_size),
        ):
            batch = [x[0] for x in batch_iterator]
            step(batch)
            yield from batch

    return f


def batch(step: Callable[[list[V]], Iterable[U]], batch_size: int) -> Step[V, U]:
    def f(data: Iterable[V]) -> Iterable[U]:
        for _, batch_iterator in groupby(
            zip(data, count()),
            key=lambda x: math.floor(x[1] / batch_size),
        ):
            yield from step([x[0] for x in batch_iterator])

    return f


@overload
def filter(step: Callable[[V], TypeGuard[W]]) -> Step[V, W]:
    ...


@overload
def filter(step: Callable[[V], bool]) -> Step[V, V]:
    ...


def filter(step: Callable[[V], bool]) -> Step[V, V]:  # type: ignore
    return partial(raw_filter, step)  # type: ignore


@overload
def branch(
    step1: Step[T, U] | None,
    step2: Step[T, V],
    pick_first: Literal[True],
    max_inflight: int | None = ...,
) -> Step[T, U]:
    ...


@overload
def branch(
    step1: Step[T, U],
    step2: Step[T, V],
    pick_first: Literal[False] | None,
    max_inflight: int | None = ...,
) -> Step[T, V | U]:
    ...


@overload
def branch(
    step1: Step[T, U],
    step2: Step[T, V],
    step3: Step[T, W],
    pick_first: Literal[False] | None,
    max_inflight: int | None = ...,
) -> Step[T, V | U | W]:
    ...


@overload
def branch(  # noqa: PLR0913
    step1: Step[T, U],
    step2: Step[T, V],
    step3: Step[T, W],
    step4: Step[T, X],
    pick_first: Literal[False] | None,
    max_inflight: int | None = ...,
) -> Step[T, V | U | W | X]:
    ...


@overload
def branch(
    *steps: Step[T, Any] | None,
    pick_first: Literal[False] | None,
    max_inflight: int | None = ...,
) -> Step[T, Any]:
    ...


def branch(  # type: ignore
    *steps: Step[T, Any] | None,
    max_inflight: int = 1000,
    pick_first: bool = False,
) -> Step[T, Any]:
    """
    Returns a step that forks the input iterable into multiple iterables,
    each one being processed by a different step. The output iterable is the
    concatenation of the output of each step.

    If `pick_first` is True, the output iterable is the concatenation of the
    output of the first step only.
    """

    def wrapper(iterable: Iterable[T]) -> Iterable[Any]:
        it = iter(iterable)

        queues: list[deque[T]] = [deque() for _ in steps]
        # could be rewritten with a single deque to be more memory efficient

        # the set of iterators that are paused because we have too many inflight items
        # they should be resumed when the number of inflight items goes down
        paused_iterators: set[int] = set()

        def gen(i: int) -> Iterator[T]:
            while True:
                mydeque = queues[i]
                if not mydeque:  # when the current deque is empty
                    try:
                        newval = next(it)  # fetch a new value and
                    except StopIteration:
                        return
                    for d in queues:  # load it to all the deques
                        d.append(newval)

                    # if there are too many inflight items, pause the iterator
                    nb_inflights = sum(len(q) for q in queues)
                    if nb_inflights > max_inflight:
                        paused_iterators.add(i)
                        return

                yield mydeque.popleft()

        iterators = [iter((steps[i] or identity)(gen(i))) for i in range(len(steps))]

        # the set of iterators that are not done yet
        pending_iterators = set(range(len(iterators)))

        while len(pending_iterators):
            i = max(  # the index of the iterator with the most inflight items
                pending_iterators,
                key=lambda i: len(queues[i]),
            )
            try:
                val = next(iterators[i])
                if not pick_first or i == 0:
                    yield val
            except StopIteration:
                if i in paused_iterators:  # resume the iterator
                    iterators[i] = iter((steps[i] or identity)(gen(i)))
                    paused_iterators.remove(i)
                else:
                    pending_iterators.remove(i)

    return wrapper


def identity(item: W) -> W:
    return item



from __future__ import annotations

from collections import deque
from collections.abc import Callable, Iterator
from typing import Any, Generic, Iterable, TypeGuard, TypeVar, overload

from iter_pipes.functional import (
    batch,
    branch,
    filter,
    for_batch,
    for_each,
    identity,
    map,
)

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")
X = TypeVar("X")
Y = TypeVar("Y")

__all__ = ["Pipeline", "PipelineFactory"]


raw_filter = filter


Step = Callable[[Iterable[T]], Iterable[U]]


def compose_steps(step1: Step[T, U] | None, step2: Step[U, V]) -> Step[T, V]:
    if step1 is None:
        return step2  # type: ignore

    def composed(items: Iterable[T]) -> Iterable[V]:
        return step2(step1(items))

    return composed


class IterableWrapper(Generic[T]):
    def __init__(self, iterable: Iterable[T]):
        self._iterable = iterable

    def __iter__(self) -> Iterator[T]:
        return iter(self._iterable)

    def consume(self) -> None:
        deque(self._iterable)

    def to_list(self) -> list[T]:
        return list(self._iterable)


class Pipeline(Generic[T, U]):
    step: Step[T, U] | None
    items: Iterable[T] | None

    def __init__(
        self,
        step: Step[T, U] | None = None,
        items: Iterable[T] | None = None,
    ):
        self.step = step
        self.items = items

    def for_each(self, step: Callable[[U], Any]) -> Pipeline[T, U]:
        return self | for_each(step)

    def map(self, step: Callable[[U], W]) -> Pipeline[T, W]:
        return self | map(step)

    def pipe(self, step: Step[U, V]) -> Pipeline[T, V]:
        return Pipeline(compose_steps(self.step, step), self.items)

    def for_batch(
        self, step: Callable[[list[U]], Any], batch_size: int
    ) -> Pipeline[T, U]:
        return self | for_batch(step, batch_size)

    def batch(
        self, step: Callable[[list[U]], Iterable[V]], batch_size: int
    ) -> Pipeline[T, V]:
        return self | batch(step, batch_size)

    @overload
    def filter(self, step: Callable[[U], TypeGuard[W]]) -> Pipeline[T, W]:
        ...

    @overload
    def filter(self, step: Callable[[U], bool]) -> Pipeline[T, U]:
        ...

    def filter(self, step):  # type: ignore
        return self | filter(step)  # type: ignore

    def filter_not_none(self: Pipeline[T, X | None]) -> Pipeline[T, X]:
        return self | filter(lambda item: item is not None)  # type: ignore

    @overload
    def branch(
        self,
        f1: Callable[[Pipeline[U, U]], Pipeline[U, W]],
        max_inflight: int = ...,
    ) -> Pipeline[U, W]:
        ...

    @overload
    def branch(
        self,
        f1: Callable[[Pipeline[U, U]], Pipeline[U, V]],
        f2: Callable[[Pipeline[U, U]], Pipeline[U, W]],
        max_inflight: int = ...,
    ) -> Pipeline[U, W | V]:
        ...

    @overload
    def branch(
        self,
        f1: Callable[[Pipeline[U, U]], Pipeline[U, V]],
        f2: Callable[[Pipeline[U, U]], Pipeline[U, W]],
        f3: Callable[[Pipeline[U, U]], Pipeline[U, X]],
        max_inflight: int = ...,
    ) -> Pipeline[U, W | V | X]:
        ...

    @overload
    def branch(  # noqa W291
        self,
        f1: Callable[[Pipeline[U, U]], Pipeline[U, V]],
        f2: Callable[[Pipeline[U, U]], Pipeline[U, W]],
        f3: Callable[[Pipeline[U, U]], Pipeline[U, X]],
        f4: Callable[[Pipeline[U, U]], Pipeline[U, Y]],
        max_inflight: int = ...,
    ) -> Pipeline[U, W | V | X | Y]:
        ...

    def branch(  # type: ignore
        self,
        *functions: Callable[[Pipeline[U, U]], Pipeline[U, Any]],
        max_inflight: int = 1000,
    ) -> Pipeline[U, Any]:
        steps = [f(Pipeline()).step or identity for f in functions]
        return self | branch(*steps, max_inflight=max_inflight, pick_first=False)  # type: ignore

    def branch_off(
        self,
        *functions: Callable[[Pipeline[U, U]], Pipeline[U, Any]],
        max_inflight: int = 1000,
    ) -> Pipeline[T, U]:
        steps = [f(Pipeline()).step or identity for f in functions]
        return self | branch(
            identity, *steps, max_inflight=max_inflight, pick_first=True
        )  # type: ignore

    def process(self, items: Iterable[T] | None = None) -> IterableWrapper[U]:
        input_ = items or self.items
        if not input_:
            raise ValueError("input is None")
        if not self.step:
            raise ValueError("step is None")
        return IterableWrapper(self.step(input_))

    def __call__(self, items: Iterable[T] | None = None) -> IterableWrapper[U]:
        return self.process(items)

    def __or__(self, step: Step[U, V]) -> Pipeline[T, V]:
        return self.pipe(step)


class PipelineFactory(Generic[V], Pipeline[V, V]):
    pass
